;; Generated package description from svg.el  -*- no-byte-compile: t -*-
(define-package "svg" "1.1.0.20220101.24551" "SVG image creation functions" '((emacs "25")) :url "https://elpa.gnu.org/packages/svg.html" :authors '(("Lars Magne Ingebrigtsen" . "larsi@gnus.org") ("Felix E. Klee" . "felix.klee@inka.de")) :maintainer '("Lars Magne Ingebrigtsen" . "larsi@gnus.org") :keywords '("image"))
